#ifndef BOUNDARY_FD_H
#define BOUNDARY_FD_H 1

#include "modelingOp.h"
#include "sep_params.h"

namespace Yang {

   class boundary
   {
   public://methods
      boundary() {}
	  virtual ~boundary() {}
   };

   //calculate the wavefield at the boundary during each time step
   class boundaryFD : public boundary
   {     
      public: //data members
         
         FDModelingParams* m_pars;
      
         boundaryType m_bdType;
         bool bFreeSurf;
         float dt;

      public: //methods
         boundaryFD(): boundary() {
         }
		 virtual ~boundaryFD() {}
      virtual void update(wavFieldGroup& wfg) = 0;
	  
      virtual void initOperator(FDModelingParams* pars);
   };

   class boundaryFD2d : public boundaryFD
   {
      public: //data members
      
      //vel cube size, not wavefield shape size
      int nx,nz;
      ETYPE dx,dz,dtDdx,dtDdz;
      int offx,offz; //The offset between vel cube and wavefield cube
      const cubeData<float>* vcub; //velocity field
	  
      public: //methods
         boundaryFD2d():boundaryFD() {}
		 virtual ~boundaryFD2d() {}
      
	   virtual void initOperator(FDModelingParams* pars);
      virtual void update(wavFieldGroup& wfg){}
   }; //The boundary FD2d object can also be used as a dummy object that does not do any damping on padding layers.
   
   class boundaryFD2dAbsorb: public boundaryFD2d
   {
      public: //methods
      boundaryFD2dAbsorb():boundaryFD2d() {}
         
		virtual ~boundaryFD2dAbsorb() {}
      
      virtual void initOperator(FDModelingParams* pars);
      virtual void update(wavFieldGroup& wfg);
   };
}

#endif

