#include "stdafx.h"

#include "modelingOpOneway.h"
#include "velFD2d.h"
#include <sstream>
#include "tagHack.h"
#include "cubeAlgo2.h"
#include "misc.h"

#include "dmHeaderReader.h"

using namespace Yang;


std::string sep_par_get_string(const char *arg) {
  FUNC_TRACE_BEG(sep_par_get_string);
  char x[1024];
  std::string ret;
  int iret=getch(arg,"s",x);
  if(iret==0) {
    prt("\n !!error getting par %s ", arg);
    VERIFY(0);
  } else {
    ret=x;
  }
  FUNC_TRACE_END(ret);
}

static char szCmd[4096]; //Should be enough space
//static char szWemig[1024] = "/net/server/homes/sep/yang/from-other/yaxun_DSO/Bin/wem3d.x";  //Migration op
static char szWemig[1024] = "/net/server/homes/sep/yang/from-other/yaxun_weo/bin/bwi-wem3d.x";  //Migration op
//static char szWemva[1024] = "/net/server/homes/sep/yang/from-other/yaxun_DSO/Bin/wemva3d_baspix.x";  //Tomography op
static char szWemva[1024] = "/net/server/homes/sep/yang/from-other/yaxun_weo/bin/bwi-wet3d.x";  //Tomography op

void modelingOpOneway::initOperatorEx(FDModelingParams* params, action_type ac_type) {
  FUNC_TRACE_BEG(modelingOpOneway::initOperatorEx);
   this->initOperator(params, ac_type);
  FUNC_TRACE_END();
}

void modelingOpOneway::initOperator(FDModelingParams* params, action_type ac_type) {
   FUNC_TRACE_BEG(modelingOpOneway::initOperator);
   mAction = ac_type;
   m_pars = params;
   VERIFY(params->velModelShape.get_ndim()<=3); //Just allow 2D currently
   FUNC_TRACE_END();
}

struct Yang::OneWayParamTang{
  //geometry
  int ngx,nsx,ngy,nsy;
  float ogx,dgx,  osx,dsx;
  float ogy,dgy,  osy,dsy;
  
  //One-way Propagator options.
  string pdip;  string oper;
  
  //Notice this offset indicates if the data are gather in shot,receiver or CMP axes. It has NOTHING to do with
  //subsurface offset gather.
  char offset; //y or n, 
  int ntpx; //taper points
  int nws;  //number of requencies.
  int nxpad_beg,nxpad_end;int nypad_beg,nypad_end;
  
  char szGeomPar[2048];
  char szOffImgPar[256];
  char szMigPar[1028];
  char szBSCpar[256]; //???!!! What is this in Tom oper
  //int image_xmin, image_xmax;
  int nhx;  float ohx,dhx;
  int nhy;  float ohy,dhy;
 
  string csouFileName,crecFileName;
  //These files needs to be randomized, since the dimension is different from my convention
  string vmodFileName,imagFileName,bslwFileName,dslwFileName,dimgFileName;
  string vmodTag,imagTag,bslwTag,dslwTag,dimgTag;
  string logfilename;
};

void modelingOpOneway::setOneWayParam( OneWayParamTang* par) {
  FUNC_TRACE_BEG(OneWay::setOneWayParam);
  par->pdip = m_pars->pdip;
  par->oper = m_pars->oper;
  par->nxpad_beg = m_pars->nxPadOneway;
  par->nxpad_end = par->nxpad_beg;
  par->ntpx = (par->nxpad_beg);
  par->nws  = m_pars->nws;
  
  //int mShot1_ix,mShot1_iz,mShotIncX;
  //int mTrace1ix,mTrace1iz,mTraceIncX,m_nTraces;

  axis ax_z = m_pars->velModelShape.get_axis(1);
  axis ax_x = m_pars->velModelShape.get_axis(2);
  axis ax_y(1,0,1.f);
  
  if (m_pars->bShot_offset) {
    par->offset = 'y';  //input data is sorted by (shot,offset), like marine 
  } else {
    par->offset = 'n'; //Input data is sorted by (shot,receiver) like land-geometry.
  }
  par->ngx = m_pars->m_nTraces; par->nsx = m_pars->m_nShots;
  par->ogx = ax_x.get_coord(m_pars->mTrace1ix); par->dgx = ax_x.d * m_pars->mTraceIncX;
  par->osx = ax_x.get_coord(m_pars->mShot1_ix); par->dsx = ax_x.d * m_pars->mShotIncX;

  sprintf(par->szGeomPar," ngx=%d nsx=%d ogx=%f dgx=%f osx=%f dsx=%f offset=%c ",
          par->ngx,par->nsx,par->ogx,par->dgx,par->osx,par->dsx, par->offset);
  
  char szApp[256];
  if (m_pars->b3D) {
    //m_pars->velModelShape.info("m_pars->velModelShape");
    ax_y = m_pars->velModelShape.get_axis(3);
    par->nypad_beg = m_pars->nyPadOneway; par->nypad_end = m_pars->nyPadOneway;
    par->ngy = m_pars->m_nTracesY; par->nsy = m_pars->m_nShotsY; 
    par->ogy = ax_y.get_coord(m_pars->mTrace1iy); par->dgy = ax_y.d * m_pars->mTraceIncY;
    par->osy = ax_y.get_coord(m_pars->mShot1_iy); par->dsy = ax_y.d * m_pars->mShotIncY;
    sprintf(szApp," ngy=%d nsy=%d ogy=%f dgy=%f osy=%f dsy=%f ",
            par->ngy, par->nsy, par->ogy, par->dgy, par->osy, par->dsy);
    strcat(par->szGeomPar, szApp);
  }
  
  
  bool bDoOffset = m_pars->mOffImgParams.bDoOffsetGather;
  par->nhx = 1; par->ohx = 0.0f; par->dhx = 0.0f; float d=0.f;
  if (bDoOffset) {
    d = ax_x.d; //dx
    par->ohx = m_pars->mOffImgParams.ohx * d;
    par->dhx = m_pars->mOffImgParams.dhx * d;
    par->nhx = m_pars->mOffImgParams.nhx;
    sprintf(par->szOffImgPar, " ohx=%f dhx=%f nhx=%d ", par->ohx, par->dhx, par->nhx);
    if (m_pars->b3D) {
      d = ax_y.d; //dy
      par->ohy = m_pars->mOffImgParams.ohy * d;
      par->dhy = m_pars->mOffImgParams.dhy * d;
      par->nhy = m_pars->mOffImgParams.nhy;
      sprintf(szApp, " ohy=%f dhy=%f nhy=%d ", par->ohy, par->dhy, par->nhy);
      strcat(par->szOffImgPar, szApp);
    }
  } else { //not do offset
  }
  
  std::string szAddMigPar(" nvrf=2 ");
  sprintf(par->szMigPar,"pdip=%s ntpx=%d nws=%d oper=%s nxpad_beg=%d nxpad_end=%d offset=%c %s",
          par->pdip.c_str(),par->ntpx,par->nws,par->oper.c_str(),
          par->nxpad_beg,par->nxpad_end, par->offset, szAddMigPar.c_str());

  if (m_pars->b3D) {
    char szt[256];
    sprintf(szt," nypad_beg=%d nypad_end=%d ", par->nypad_beg,par->nypad_end);
    strcat(par->szMigPar,szt);
  }
  
  //sprintf(par->szBSCpar," typbsc=2 nopbsc=3 ");
  sprintf(par->szBSCpar," bsctyp=fourie "); //new code
  
  //get the filename corresponds to the source signature.
  par->csouFileName = m_pars->csouFileName; par->crecFileName = m_pars->crecFileName;
  
  //Make randomize name for other filenames
  srand(time(NULL));
  int rsuf = 0; rsuf = rand()/65536;
  std::ostringstream oss;
  oss.str("");
  oss << "rrvmod" << rsuf <<".H"; par->vmodFileName = oss.str();
  oss << "rrvmodTag" << rsuf;     par->vmodTag = oss.str();
  oss.str("");
  oss << "rrbslw" << rsuf <<".H"; par->bslwFileName = oss.str();
  oss << "rrbslwTag" << rsuf; par->bslwTag   = oss.str();
  oss.str("");
  oss << "rrimag" << rsuf <<".H"; par->imagFileName = oss.str();
  oss << "rrimagTag" << rsuf; par->imagTag = oss.str();
  oss.str("");
  oss << "rrdimg" << rsuf <<".H"; par->dimgFileName = oss.str();
  oss << "rrdimgTag" << rsuf; par->dimgTag = oss.str();
  oss.str("");
  oss << "rrdslw" << rsuf <<".H"; par->dslwFileName = oss.str();
  oss << "rrdslwTag" << rsuf;     par->dslwTag = oss.str();
  oss.str("");
  oss << "log.oneway" << rsuf; par->logfilename = oss.str();
  FUNC_TRACE_END();
}

//Handles the format of my velocity cube and Oneway code's input velocity model
//This works for both 2D/3D cases.
void modelingOpOneway::convertVelModel(cubeData<float> &vmodYaxun, cubeData<float>& myVelCube, bool bFromMyvelToYaxunVel) {
  FUNC_TRACE_BEG(modelingOpOneway::convertVelModel);
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);

  //bool b3D = (m_pars->b3D); //It happens that 2D/3D does not change the code.
  
  cubeData<float> vmodCube;
  axis ax_z,ax_znew;
  ax_z = m_pars->velModelShape.get_axis(1); ax_znew = ax_z.alter(shotOffZ);
  vecAxis axes = m_pars->velModelShape.get_axes(); axes[0] = ax_znew;
  bool b3D = axes.size()==3;
  //LOG('h1');
  if (bFromMyvelToYaxunVel) {
    //set size vmod and copy to m_pars->pVel
    hypercube velCropShape(axes); vmodCube.alloc(velCropShape);
    cropAlign(vmodCube,myVelCube,false);
    //Then transpose from (z,x)/(z,x,y) to (x,y,z), 2D/3D are the same
    int t1[2] = {1,3}; int t2[2] = {2,1};
    stdVecInt reshp(t1,t1+2),mapp(t2,t2+2);
    shuffleAxis(vmodYaxun,vmodCube,reshp,mapp,true,true);
  } else { //convert the otherway around
    //first transpose vmodYaxun 
    //convert (x,y,z) back to (z,x,/y);
    int t1[3] = {1,2,3}; int t2[3] = {2,3,1};
    stdVecInt reshp(t1,t1+3),mapp(t2,t2+3);
    shuffleAxis(vmodCube,vmodYaxun,reshp,mapp,true,true);
    if (!b3D) {
      //strip out the last axis y.
      hypercube shape2 = vmodCube.domain().return_compatible(1,2);
      vmodCube.reshape(shape2,NULL);
    }
    cropAlign(myVelCube,vmodCube,false);
  }
  //LOG('h2');
  vmodYaxun.dump("dump1");
  FUNC_TRACE_END();
}


//This works only for 2D case, where my img is (z,x,hx) and yaxun's img is (hx,hy,x,y,z)
void modelingOpOneway::convertImage(cubeData<float> &ImageYaxun, cubeData<float>& myImgCube, bool bFromMyImageToYaxun) {
  FUNC_TRACE_BEG(modelingOpOneway::convertImage);
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);

  cubeData<float> imgN;
  axis ax_z,ax_znew;
  ax_z = myImgCube.domain().get_axis(1); ax_znew = ax_z.alter(shotOffZ);
  vecAxis axes = myImgCube.domain().get_axes(); axes[0] = ax_znew;
  hypercube myImgShape(axes);
  bool bDoOffsetGather = m_pars->mOffImgParams.bDoOffsetGather;
  stdVecInt reshp,mapp;
  
  if (bFromMyImageToYaxun) {
    //Crop my image first
    imgN.alloc(myImgShape); 
    cropAlign(imgN,myImgCube,false);
    //flip axes from (z,x,hx,...hy,y) to (hx,hy,x,y,z) 
    reshp.resize(4); reshp[0]=1; reshp[1]=2; reshp[2]=4; reshp[3]=5;
    mapp.resize(4);  mapp[0] =4; mapp[1] =2; mapp[2] =1; mapp [3]=3;
    Yang::shuffleAxis(ImageYaxun,imgN,reshp,mapp,true,true);
  } else {
    //Arrange the image in axis (hx,hy,x,y,z) to (z,x) or (z,x,hx)
    reshp.resize(4); reshp[0]=2; reshp[1]=3; reshp[2]=4; reshp[3]=5;
    mapp.resize(4);  mapp[0] =3; mapp[1] =2; mapp[2] =4; mapp [3]=1;
    Yang::shuffleAxis(imgN,ImageYaxun,reshp,mapp,true,true);
    //Clip the last axis
    hypercube shape2;
    if (!bDoOffsetGather) {
      shape2 = imgN.domain().return_compatible(1,2);
    } else {
      shape2 = imgN.domain().return_compatible(1,3);
    }
    imgN.reshape(shape2);
    //Now crop imgN to pImageCube;
    cropAlign(myImgCube,imgN,false);
  }
  //imgN.info("imgN info in convertImage Oneway");
  FUNC_TRACE_END();
}

//Yaxun's image has the ordering of (hx,hy,x,y,z), my 3D ordering would be (z,x,y,hx,hy). Therefore Transp <myimg3d.H reshape=3,5 plane=12 | Transp reshape=2,3,5 plane=23 > yaxunimg3d.H
//This should work for both 2D/3D cases.
void modelingOpOneway::convertImage(const string& fn_ImageYaxun, const string& fn_myImg, bool bFromMyImageToYaxun) {
  FUNC_TRACE_BEG("modelingOpOneway::convertImage3d");
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);
  VERIFY(m_pars->b3D);
  cubeData<float> imgN;
  axis ax_z,ax_znew;
  
  const hypercube& shape0offimg = m_pars->finalImageShape;
  const hypercube& shape_offimg = m_pars->finalOffImageShape;
  
  ax_z = shape0offimg.get_axis(1); ax_znew = ax_z.alter(shotOffZ);
  //vecAxis axes = shape0offimg.get_axes(); axes[0] = ax_znew;
  //hypercube myImgShape(axes);
  bool bDoOffsetGather = m_pars->mOffImgParams.bDoOffsetGather;
  const hypercube* pMyImgCube = bDoOffsetGather? &shape_offimg:&shape0offimg;
  
  stdVecInt reshp,mapp;
  bool b3D = m_pars->b3D;
  int n_axis_img = 3;
  if (b3D) n_axis_img = 5;
  if (bFromMyImageToYaxun) {
    //Then just Window and then transp the images to comply with Yaxun's format
    sepHeaderReader myImgHeader(fn_myImg,dm::IN,"",true);
    hypercube shape2;
    myImgHeader.getShape(shape2,n_axis_img); 
    PRV(fn_myImg); shape2.info("shape2"); pMyImgCube->info("image shape");
    VERIFY(shape2.sameSize(*pMyImgCube));
    if (!b3D) {
      //flip axes from (z,x,hx,...hy,y) to (hx,hy,x,y,z) 
      sprintf(szCmd, "Window3d squeeze=n f1=%d <%s | Transp reshape=2,4,5 plane=12 | Transp reshape=2,3,5 plane=23 >%s", shotOffZ, fn_myImg.c_str(), fn_ImageYaxun.c_str());
    } else {
      //flip axes from (z,x,y,hx,hy) to (hx,hy,x,y,z) 
      sprintf(szCmd, "Window3d squeeze=n f1=%d <%s | Transp reshape=1,3,5 plane=12 | Transp reshape=3,5 plane=12 >%s", shotOffZ, fn_myImg.c_str(), fn_ImageYaxun.c_str());
    } //If no offset image, then all these dummy axes (hx,hy) are initally at the end, it can be processed in the same way and there is no need to differentiate.
  } else { //from Yaxun's image to mine
    //Arrange the image in axis (hx,hy,x,y,z) to (z,x) or (z,x,hx)
    if (!b3D) {
      //flip axes from (hx,hy,x,y,z) to (z,x,hx,...hy,y) and pad
      sprintf(szCmd, " Window3d <%s squeeze=n | Transp reshape=2,3,4,5 plane=12 | Transp reshape=4,5 plane=12 | Pad beg1=%d >%s", fn_ImageYaxun.c_str(), shotOffZ, fn_myImg.c_str());
    } else {
      //flip axes from (hx,hy,x,y,z) to (z,x,y,hx,hy) and pad z
      sprintf(szCmd, " Window3d <%s squeeze=n | Transp reshape=2,5 plane=12 | Transp reshape=2,3,5 plane=12 | Pad beg1=%d >%s",     fn_ImageYaxun.c_str(), shotOffZ, fn_myImg.c_str());
    }
  } //end if bFromMyImageToYaxun
  prt("Executing Cmd in modelingOpOneway convertImage 3D:%s \n", szCmd);
  VERIFY(system(szCmd)==0);
  FUNC_TRACE_END();
}


//No non-lin modeling for one-way equation
void modelingOpOneway::modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec) {
  FUNC_TRACE_BEG(modelingOpOneway::modeling_op);
  VERIFY(0);
  FUNC_TRACE_END()
}

//Invoke Yaxun's oneway program to perform tasks.
//Invoke wem3d.x, not finished yet..............
void modelingOpOneway::bornModeling_op(cubeData<float>* pReflCube, dmRsf<float>* pRsfSeisRec) {
  OneWayParamTang param;
  this->setOneWayParam(&param);
  //!!!Need to set the depth of the source, what param?;
  //First check the size is correct or not. And align the offset correspondingly.
  //Then transpose the input
  //Dump the input to file rrimag%
  //invoke the wem3d command, after then the result is in crec file
  //use another dmRsf_oc object to read the crec file, and read in the content
  //and update the pRsfSeisRec content.
  VERIFY(0);
}

//Input changed to a tag string
void modelingOpOneway::migration_op(const string& seis_fn, cubeData<float>* pImageCube) {
  FUNC_TRACE_BEG("modelingOpOneway::migration_op2");
  //Get file name from its tag
  //string crecFileName = sep_par_get_string(dataTag.c_str());
  OneWayParamTang param;
  this->setOneWayParam(&param);
  param.crecFileName = seis_fn;
  
  //Do the actual migration
  this->migration(param,pImageCube);

  //If sparse sample the image
  if (m_pars->pImageMask) {
    pImageCube->multiply(*(m_pars->pImageMask));
  }
  FUNC_TRACE_END();
}

void modelingOpOneway::migration_op(dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube) {
  FUNC_TRACE_BEG("modelingOpOneway::migration_op");
  //Figure out the data filename, assume its data file never changes (valid for WEMVA but not for FWI, while one-way is not used in FWI anyway, so it is safe to do this);
  VERIFY(!pRsfSeisRec->isDummy() && pRsfSeisRec->isInput());
  //Get file name from its tag
  string dTag = pRsfSeisRec->getTag();
  this->migration_op(dTag, pImageCube);
  FUNC_TRACE_END();
}

//!!!Need to set the depth of the source, what param?; Ask Yaxun
//First check the size is correct or not. And align the offset correspondingly.
//Check if the input pRsfSeisrec is of correct shape, then 
//if RsfSeisrec is a dmRsf_oc, then take its tag to find corresponding filename. 
//else the let pRsfSeisrec dump to "crec" tag.
//invoke the wem3d command, //direct the output to file rrimag%
//use another dmRsf_oc object to read rrimag% file, and read in the content
//The axes of image is (z,x,hx)
void modelingOpOneway::migration(OneWayParamTang& param, cubeData<float>* pImageCube) {
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);
  //also need wavelet(csou) file, get from params
  //Also need to dump the velocity model
  cubeData<float> vmodCube2;

  //set size vmod and copy to m_pars->pVel, convert my velmodel to Yaxun
  this->convertVelModel(vmodCube2,*(m_pars->pVel),true);
  //Dump the vmod
  //addNewSepTag(param.vmodTag.c_str(),false,param.vmodFileName.c_str()); vmodCube2.dump(param.vmodTag,true); 
  //Alternativly, we can just dump to the filename
  vmodCube2.dump(param.vmodFileName,true);  usleep(int(1e5)); //for IO safety

  //LOG(Now the inputs are ready);
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgadj memchk=n report=y %s %s >&%s", //redirect the output
          szWemig,param.csouFileName.c_str(),param.crecFileName.c_str(),param.vmodFileName.c_str(),
          param.imagFileName.c_str(),param.szMigPar,param.szOffImgPar, param.logfilename.c_str());
  PRV(szCmd);
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //LOG(finish invoking the wem3d program);

  //Need read the result image back
  cubeData<float> imgY,imgN;
  readSepFileUsingFileName(imgY,param.imagTag,param.imagFileName); usleep(int(1e5));
  this->convertImage(imgY,*pImageCube,false);

  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.vmodFileName.c_str(),param.imagFileName.c_str());
  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
}

//For 3D case
void modelingOpOneway::migration3d_op(const string& fn_seisRec, const string& fn_img_off) {
  FUNC_TRACE_BEG(modelingOpOneway::migration3d_op);
  //Figure out the data filename, assume its data file never changes (valid for WEMVA but not for FWI, while one-way is not used in FWI anyway, so it is safe to do this);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  param.crecFileName = fn_seisRec;
  
  //Do the actual migration
  this->migration3d(param,fn_img_off);
  
  //If sparse sample the image
  char szFnt[256]; sprintf(szFnt,"junk%d.H",rand()%256);
  if (!(m_pars->fn_imageMask.empty())) {
    sprintf(szCmd, " Add %s %s mode=prod >%s " ,m_pars->fn_imageMask.c_str(),fn_img_off.c_str(), szFnt);
    PRV(szCmd); VERIFY(system(szCmd)==0);
    sprintf(szCmd, " cp %s %s", szFnt, fn_img_off.c_str());
  }
  FUNC_TRACE_END();
}

//for 3D case.
void modelingOpOneway::migration3d(OneWayParamTang& param, const string& fn_img_off) {
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ); //One-way code assume the src receiver on the same altitude.
  //also need wavelet(csou) file, get from params
  //Also need to dump the velocity model
  cubeData<float> vmodCube2; //Yaxun's vel model

  //Convert my velocity model to Yaxun's vel model
  //set size vmod and copy to m_pars->pVel, crop velocity model.
  this->convertVelModel(vmodCube2,*(m_pars->pVel), true);
  
  axis ax_z,ax_znew;
  ax_z = m_pars->velModelShape.get_axis(1); ax_znew = ax_z.alter(shotOffZ);
  vecAxis axes = m_pars->velModelShape.get_axes(); axes[0] = ax_znew;

  //we can just dump the new velocity model to the filename
  vmodCube2.dump(param.vmodFileName,true); 
  usleep(int(1e5)); //for IO safety

  //LOG(Now the inputs are ready);
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s bimg=%s mode=imgadj memchk=n report=y %s %s >& %s", //redirect the output
          szWemig,param.csouFileName.c_str(),param.crecFileName.c_str(),param.vmodFileName.c_str(),
          param.imagFileName.c_str(),param.szMigPar,param.szOffImgPar, param.logfilename.c_str());
  prt("Executing cmd:%s. \n",szCmd);
  VERIFY(system(szCmd)==0); usleep(int(1e5));

  //LOG(finish invoking the wem3d program);
  //Need now convert the result image back to my own format.
  this->convertImage(param.imagFileName, fn_img_off, false);
  
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s; Rm %s",param.vmodFileName.c_str(),param.imagFileName.c_str());
  //sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
}
  
//Input Image perturbation and output slowness perturbation
void modelingOpOneway::TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS){
  FUNC_TRACE_BEG(modelingOpOneway::TomAdj_op);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);
  
  //Dump back ground slowness
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,*(m_pars->pVel),true);
  vmodCube2.reciprocal(); //convert to slowness
  addNewSepTag(param.bslwTag.c_str(),false,param.bslwFileName.c_str());
  vmodCube2.dump(param.bslwTag,true); usleep(int(1e5));
  
  //Convert the image perturbation
  cubeDataF Image2;
  this->convertImage(Image2,*pDeltaIh,true);
  addNewSepTag(param.dimgTag.c_str(),false,param.dimgFileName.c_str());
  Image2.dump(param.dimgTag,true); usleep(int(1e5));
  
  //LOG(Now the inputs are ready to do Tom_adj);
  sprintf(szCmd,"%s csou=%s crec=%s bslw=%s dslw=%s dimg=%s adj=y %s %s %s >&%s",
          szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bslwFileName.c_str(),
          param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szBSCpar,param.szOffImgPar,param.logfilename.c_str());
  PRV(szCmd);
  
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //LOG(finish invoking the wemvaTom program);
  
  //Need read the result image back
  cubeData<float> slowY;
  slowY.load(param.dslwFileName,3); usleep(int(1e5));
  
  //Now convert shape from slowY to pDeltaS
  this->convertVelModel(slowY,*pDeltaS,false);
  
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bslwFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

//Input Image perturbation and output slowness perturbation, 3D case.
void modelingOpOneway::TomAdj3d_op (const string& fn_dImg_off, cubeData<float>& dslwCube){
  FUNC_TRACE_BEG("modelingOpOneway::TomAdj3d_op");
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);
  
  //Dump back ground slowness
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,*(m_pars->pVel),true);
  //vmodCube2.reciprocal(); //convert to slowness
  //addNewSepTag(param.bslwTag.c_str(),false,param.bslwFileName.c_str());
  //vmodCube2.dump(param.bslwTag,true); usleep(int(1e5));
  vmodCube2.dump(param.bslwFileName,true); usleep(int(1e5)); //!! Notice it is velocity now
  
  //Convert the image perturbation to Yaxun's format
  this->convertImage(param.dimgFileName,fn_dImg_off,true);
  
  //LOG(Now the inputs are ready to do Tom_adj);
  sprintf(szCmd,"%s csou=%s crec=%s bvel=%s dvel=%s dimg=%s adj=y nvrf=2 memchk=n report=y mode=tomadj %s %s %s >&%s",
          szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bslwFileName.c_str(),
          param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szBSCpar,param.szOffImgPar,param.logfilename.c_str());
  PRV(szCmd);
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  //LOG(finish invoking the wemvaTom program);
  
  LOGC("sdf1");
  //Need to convert the result slowness model back
  cubeDataF dslwYaxun;  dslwYaxun.load(param.dslwFileName,3);
  dslwCube.domain().info("dslwCube");
  this->convertVelModel(dslwYaxun,dslwCube,false);
  
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bslwFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  //sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}

void modelingOpOneway::TomFwd_op (cubeData<float>* pDeltaS, cubeData<float>* pDeltaI){
  
  FUNC_TRACE_BEG(modelingOpOneway::TomFwd_op);
  OneWayParamTang param;
  this->setOneWayParam(&param);
  
  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first  
  int shotOffZ = m_pars->mTrace1iz;
  VERIFY(m_pars->mShot1_iz == shotOffZ);
  
  //Dump back ground slowness
  cubeDataF vmodCube2;
  this->convertVelModel(vmodCube2,*(m_pars->pVel),true);
  vmodCube2.reciprocal(); //convert to slowness
  addNewSepTag(param.bslwTag.c_str(),false,param.bslwFileName.c_str());
  vmodCube2.dump(param.bslwTag,true); usleep(int(1e5));
  
  //Convert the slowness perturbation
  cubeDataF slow2;
  this->convertVelModel(slow2,*pDeltaS,true);
  addNewSepTag(param.dslwTag.c_str(),false,param.dslwFileName.c_str());
  slow2.dump(param.dslwTag,true); usleep(int(1e5));
  
  LOG("Now the inputs are ready to do Tom_adj");
  sprintf(szCmd,"%s csou=%s crec=%s bslw=%s dslw=%s dimg=%s adj=n %s %s %s >&%s",
          szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bslwFileName.c_str(),
          param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szBSCpar,param.szOffImgPar,param.logfilename.c_str());
  PRV(szCmd);
  
  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
  LOG("finish invoking the wemvaTom program");

  //Need read the result image back
  cubeData<float> imageY;
  readSepFileUsingFileName(imageY,param.dimgTag,param.dimgFileName); usleep(int(1e5));
  
  //Now convert shape from slowY to pDeltaS
  this->convertImage(imageY,*pDeltaI,false);
  
  //Remove the intemediate files
  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bslwFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
  verbPrt("Exec cmd:%s\n",szCmd);
  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
  FUNC_TRACE_END();
}


///// Not finished yet! ~~~!!
//void modelingOpOneway::TomFwd3d_op (cubeData<float>* pDeltaS, cubeData<float>* pDeltaI){
//
//  FUNC_TRACE_BEG(modelingOpOneway::TomFwd_op);
//  OneWayParamTang param;
//  this->setOneWayParam(&param);
//
//  //The source/receiver is always assumed to be on top of the velocity model, Thus crop the velocity model first
//  int shotOffZ = m_pars->mTrace1iz;
//  VERIFY(m_pars->mShot1_iz == shotOffZ);
//
//  //Dump back ground slowness
//  cubeDataF vmodCube2;
//  this->convertVelModel(vmodCube2,*(m_pars->pVel),true);
//  vmodCube2.reciprocal(); //convert to slowness
//  addNewSepTag(param.bslwTag.c_str(),false,param.bslwFileName.c_str());
//  vmodCube2.dump(param.bslwTag,true); usleep(int(1e5));
//
//  //Convert the slowness perturbation
//  cubeDataF slow2;
//  this->convertVelModel(slow2,*pDeltaS,true);
//  addNewSepTag(param.dslwTag.c_str(),false,param.dslwFileName.c_str());
//  slow2.dump(param.dslwTag,true); usleep(int(1e5));
//
//  LOG("Now the inputs are ready to do Tom_adj");
//  sprintf(szCmd,"%s csou=%s crec=%s bslw=%s dslw=%s dimg=%s adj=n %s %s %s >&%s",
//          szWemva,param.csouFileName.c_str(),param.crecFileName.c_str(),param.bslwFileName.c_str(),
//          param.dslwFileName.c_str(),param.dimgFileName.c_str(),param.szMigPar,param.szBSCpar,param.szOffImgPar,param.logfilename.c_str());
//  PRV(szCmd);
//
//  int sysret = system(szCmd); VERIFY(sysret==0); usleep(int(1e5));
//  LOG("finish invoking the wemvaTom program");
//
//  //Need read the result image back
//  cubeData<float> imageY;
//  readSepFileUsingFileName(imageY,param.dimgTag,param.dimgFileName); usleep(int(1e5));
//
//  //Now convert shape from slowY to pDeltaS
//  this->convertImage(imageY,*pDeltaI,false);
//
//  //Remove the intemediate files
//  sprintf(szCmd,"Rm %s;Rm %s;Rm %s",param.bslwFileName.c_str(),param.dimgFileName.c_str(),param.dslwFileName.c_str());
//  sysret = system(szCmd); assert(sysret==0); usleep(int(1e5));
//  FUNC_TRACE_END();
//}

