#include "stdafx.h"
#include "param_func.h"
#include "param_clib.h"
#include "solverNonlinearSearch.h"
#include "nonLinSolver.param.h"

using namespace Yang;

//Read nonLinSolver Parameters
void Yang::read_nonLinSolver_par(solverParam* pSolvPar, param_func* pars) {
  solverParam& solvPar = *pSolvPar;
  bool bDoSmooth = pars->get_bool("bDoSmooth",false);
  if (bDoSmooth) {
    int smTypeIndex = pars->get_int("smType",0);
    switch (smTypeIndex) {
    case 0:
    case 1:
      if (smTypeIndex==0) {solvPar.eSmooth = solverParam::sm_spline;}
      else {
        solvPar.eSmooth = solverParam::sm_rect; 
        prt("!Use seplib Rect Smooth!\n");
      }
      //0 means no smoothing on the model
      solvPar.nSpl[0]  = pars->get_int("n1spl",0);
      solvPar.nSpl[1]  = pars->get_int("n2spl",0);
      solvPar.nSpl[2]  = pars->get_int("n3spl",0);
      break;
    case 2:
      solvPar.eSmooth = solverParam::sm_alongDip; 
      VERIFY(0);
      break;
    default:
      VERIFY(0);
    }
    solvPar.repeat  = pars->get_int("repeat",1);
  }
  
  solvPar.bSmoothSearchDir = bDoSmooth;
  solvPar.bChangeSmoothParamWithIteration = pars->get_bool("bVarySmoothing",false);
  float fmaxVel        = pars->get_float("fmaxVel");//,3200.0f); //max allowed vel
  float fminVel        = pars->get_float("fminVel");//,1500.0f); //min allowed vel
  PRV(fmaxVel); PRV(fminVel);
  solvPar.fmaxValAllow = 1.0f/fminVel; //slowness
  solvPar.fminValAllow = 1.0f/fmaxVel;
  solvPar.maxIter      = pars->get_int("maxNonLinIter",5);
  solvPar.errTol       = pars->get_float("errorAllowed",1e-4);
  solvPar.fMaxPerturbScale = pars->get_float("maxPerturbScale",0.05); //5 percent of the slowness RMS value
  solvPar.bDecreaseMaxStepLengthWithIteration = pars->get_bool("bVaryStepLength",false);
  solvPar.fbase1       = pars->get_float("maxPerturbBase",0.2f);
  solvPar.fdecay       = pars->get_float("maxPerturbDecay",0.8f);
  
  //Gradient Mask
  bool bMaskGrad = pars->get_bool("bMaskGradient",false);
  solvPar.bMaskGradient = bMaskGrad;
  bool b3d = pars->get_bool(csz_b3D,false);
  if (bMaskGrad) { //load gradmask
    if (!b3d) {
      solvPar.gradMask.load("gradMask",2);
    } else {
      solvPar.gradMask.load("gradMask",3); //as 3D cube.
    }
    solvPar.gradMask.info("gradMask");
  }
}
