#ifndef STDAFX_H
#define STDAFX_H 1

#include "stdafxc.h"

#include <iostream>
#include <stdexcept>
//#include <direct.h>
#include <vector>
#include <algorithm>
#include <numeric>
#include <memory>
#include <stdio.h>


#include <typeinfo>

//OpenMP support
#include <omp.h>
#include "myOMP.h"

#include <iostream>
#include <assert.h>
#include <valarray>
#include <complex>
using std::cout;
using std::cerr;
using std::endl;
using std::vector;
using std::string;
using std::auto_ptr;
using std::valarray;

typedef std::vector<float> stdVecFloat;
typedef std::vector<int> stdVecInt;
typedef std::vector<double> stdVecDb;
typedef std::vector<float> stdvecFloat;
typedef std::vector<int> stdvecInt;
typedef std::vector<double> stdvecDb;

//MY LIB
#include "axis.h"
#include "hypercube.h"

//Macros

#ifndef SAFE_DELETE
#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#endif    
#ifndef SAFE_DELETE_ARRAY
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }
#endif    
#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p)=NULL; } }
#endif

#define ITER_SOLVER_TOL   (1e-6f)
//constants
const double PI=3.14159265358979323846264338327950288419716939937510;
const float epsilon=1e-7f;

typedef std::complex<float>   cmplxf;
typedef std::complex<double>  cmplxd;

template <typename T>
inline T* va_ptr(std::valarray<T>& va) {return &(va[0]);}

template <typename T>
void valarray_assign(valarray<T>& array, const T* array_c) {
  for (ulong i=0; i<array.size(); i++) {
    array[i] = array_c[i];
  }
}

template <typename T>
void valarray_assignr(T* array_c, const valarray<T>& array) {
  for (ulong i=0; i<array.size(); i++) {
    array_c[i]  = array[i];
  }
}

//Decide float number equal or not
template <typename T>
bool equal(const T& x, const T& y) 
{
   return (fabs(x-y)<epsilon);
}

bool equalf(const float& x,const float& y);

float rad2deg(float rad);
float deg2rad(float angle);

//Taking SQUARES
template <typename T>
T SQ(const T& a) {return a*a;}
template <typename T>
T CB(const T& b) {return b*b*b;}

//Already defined in SEPINC headers
//#define MIN(a,b)  ( (a)<(b)? (a):(b) )
//#define MAX(a,b)  ( (a)>(b)? (a):(b) )

//uncomment this line to use double precision in compuation
//#define USE_DB_PRECISION
 
#ifndef USE_DB_PRECISION
   #define ETYPE float  // or double if you want
#else
   #define ETYPE double
#endif      

//For verbose message
/*#define VERB_PRT(args) {   			\
						if (gVerb)  \//verbose print
						   (printf("VERB: "), printf args); \ 
						}
*/
//Cannot use printf because most cases the stdout is redirected, need to use stderr
//#define VERB_PRT(args) {if (gVerb)  (printf("VERB: "), printf args);}

#define PRT_LOC {cerr << endl     <<__FILE__<<":"<<__LINE__<<"; ";}

#define PRV(x)  {cerr << __FILE__ <<":"<<__LINE__<<"; "<< #x "= " << x << endl;}
#define PRV2(x,y)     {PRV(x); PRV(y);}
#define PRV4(x,y,z,w) {PRV2(x,y); PRV2(z,w);}
#define PRV8(x1,x2,x3,x4,x5,x6,x7,x8) {PRV4(x1,x2,x3,x4); PRV4(x5,x6,x7,x8);}

#define LOG(word) \
	if (gIsLogging) \
		{cerr<<__FILE__<<":"<<__LINE__<<"; "<< #word << endl;}

#define PRVEC(x,n) {cerr<<__FILE__<<":"<<__LINE__<<"; "<< #x "= "; \
				    for (int k_v=0; k_v < n; k_v++) { \
					cerr << (x)[k_v] << ", "; } \
            cerr << endl; }
#define VERB_PRV(x) \
   { if(gVerb) cerr<<"\n"<<__FILE__<<":"<<__LINE__<<"; "<< #x "= " << x;} 
//Print single variable 

#ifdef DEBUG
   #define DB_PRV(x) {cerr<<__FILE__<<":"<<__LINE__<<"; "<< #x "= " << x << "\n" ;}
#else
   #define DB_PRV(x) {}
#endif

//For expression
#ifdef DEBUG
   #define TRACE(S) { cerr<< #S << endl, S;}
#else
   #define TRACE(S) {S;} 
#endif

extern std::vector<std::string> g_szVecCallStack;
extern char g_szCallStackTrc[];

//For Stack trace
#ifndef NDEBUG
#define FUNC_TRACE_BEG(word) { {sprintf(g_szCallStackTrc,"\n %s:line %d; "#word,__FILE__,__LINE__);  \
						   g_szVecCallStack.push_back(std::string(g_szCallStackTrc));}
#define FUNC_TRACE_END(ret) g_szVecCallStack.pop_back(); return ret; }

#else
#define FUNC_TRACE_BEG(word) ;
#define FUNC_TRACE_END(ret)  return ret;
#endif

//Define Modeling, RTM, adjRTM operations for GPU engine.
const char* const sep=",";

//For sep_err quick tracef, like verify macro in MFC
#define VERIFY(S) {   \
  if(!(S))   \
   {  \
      cerr <<"\n"<<__FILE__<<":"<<__LINE__<<"; "<<"!!(" <<  #S << ") Did not pass verify !!";\
	  assert(0);\
	  abort(); \
   }} 

//Casting
#define DYN_CAST_PTR(pTarg,pSrc,derClassName) 	derClassName* pTarg = dynamic_cast<derClassName*>(pSrc)
#define DYN_CAST_REF(Targ,Src,derClassName) 	derClassName& Targ = dynamic_cast<derClassName&>(Src)
#define STATIC_CAST_PTR(pTarg,pSrc,derClassName) 	derClassName* pTarg = static_cast<derClassName*>(pSrc)
#define STATIC_CAST_REF(Targ,Src,derClassName) 	derClassName& Targ = static_cast<derClassName&>(Src)
#define CONST_CAST_PTR(pTarg,pSrc,ClassName) 	ClassName*    pTarg = const_cast<ClassName*>(pSrc)
#define CONST_CAST_REF(Targ,Src,ClassName) 	  ClassName& Targ  = const_cast<ClassName&>(Src)

//ptr is the class obj ptr, func is the func name along with the parameters)
//If the ptr is NULL, then do not execute the function
#define SAFE_PTR_CALL(ptr,func) \
  if (NULL!=ptr) { \
    (ptr)->func; \
  }
  

// Maximum 2Gigabytes 
const unsigned int MAX_MEMORY_ALLOC = 2048000000; //~2G


//Functions
//Some function wrappers, it might out of work if ElemCount is too big!!!!!
template<typename T>
inline void myMemcpy(T* dest,const T* src, size_t ElemCount) {
   T* pSrc = const_cast<T*>(src);
   memcpy((void*)(dest),(void*)(pSrc),sizeof(T)*ElemCount);
}
template<typename T>
inline void myMemset(T* dest,T val, size_t ElemCount) {
  long long i;
  for (i=0; i< ElemCount; i++) {
    *(dest+i) = val;
  }
}

template<typename T>
void prtStdVec(const std::vector<T>& vec) {
   int cnt=vec.size();
   int i;
   for (i=0; i<cnt; i++) {
	  cerr << "'" << vec[i] << "'";
   }
   cerr << endl;
}

//Calc the statistics of a numeric array.
template<typename T>
void arrayStat(const char* arrName, const T* arr, long len) {
  long index=0, nz=0;
  float fsum, fsqr,f, fmax,fmin;
  fsum= fsqr=0.0f;
  fmax = fmin = arr[0];
  for (index=0; index<len; index++) {
    f     = arr[index];
    fsum += f;
    fsqr += f*f;
    if (fmax < f) fmax=f;
    if (fmin > f) fmin=f;
    if (f!=0.f) {
      nz++;
    }
  }
  float fmean =  fsum/float(len);
  prt("arr stat of %s:size %ld, min:%5.3g max:%5.3g mean=%5.3g, sqsum=%5.3g, Nonzeros=%10li\n", 
         arrName,len,fmin,fmax,fmean,fsqr,nz);
  return;
}

#endif //include protection macro
