#ifndef __WAVEPROPAGATE_PARAM_H_INCL__
#define __WAVEPROPAGATE_PARAM_H_INCL__

// TODO: your code here
#include "modelingFD_aux.h"
#include "param_entry.h"
namespace Yang {
inline void read_twoway_prop_param(FDModelingParams* modelingPar, param_func* pars) {
  FDModelingParams& params = *modelingPar;
  
  params.b3D = pars->get_bool(csz_b3D,false); //If 3D or not.
  //also get the time param from the params
  params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
  params.stencilWidth = pars->get_int(csz_stenWidth);
  params.bUse4thOrder= false; //no 4th order in time.
  params.offx = params.stencilWidth;params.offz = params.stencilWidth;params.offy = params.stencilWidth;
  params.m_nPad = pars->get_int(csz_npad);  

  //Read in Absorption type
  int itype = pars->get_int(csz_bdType);
  switch (itype)
  {
   case 1:
      params.m_bdType = abs_ord1;
      break;
   case 2:
      params.m_bdType = abs_ord2;
      break;
   case 3:
      params.m_bdType = abs_Dave;
      break;
   default:
      seperr("!unknown boundary condition type!\n");
  }
  params.mImposeRecvType = AsForce;
  params.bTake2ndTimeDerivForRTM = pars->get_bool(csz_take2ndTimeDerivRTM,false);
  
  params.pSrcSnapshot = NULL;  params.pSrcSnapshot2 = NULL;
  params.pRecvSnapshot = NULL; params.pRecvSnapshot2 = NULL;
  
  //calc time parameters
  params.timeParam.dtComp = pars->get_float(csz_dtComp);
  params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
  params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);   
  //Always assume the recording starts from zero time. !! 
  params.timeParam.dtRecord = pars->get_float(csz_dtRec);
  params.timeParam.totalTime = pars->get_float(csz_totalTime);
}

inline void read_migration_param(FDModelingParams* pModelingPar, param_func* pars) {
  FDModelingParams& params = *pModelingPar;
  params.b3D = pars->get_bool(csz_b3D,false); //If 3D or not.
  bool b3D = params.b3D;
  
  //Load the offset Image parameters.
  bool bDoOffsetGather = pars->get_bool(csz_do_offset,true);
  params.mOffImgParams.bDoOffsetGather = bDoOffsetGather;
  if (bDoOffsetGather) {
    params.mOffImgParams.dhx = pars->get_int(csz_img_dhx);
    params.mOffImgParams.nhx = pars->get_int(csz_img_nhx);
    params.mOffImgParams.ohx = pars->get_int(csz_img_ohx);
    params.mOffImgParams.AxesUsage = 0x1;

    params.mOffImgParams.nhy = 1;
    if (b3D) { //need y now.
      params.mOffImgParams.dhy = pars->get_int(csz_img_dhy);
      params.mOffImgParams.nhy = pars->get_int(csz_img_nhy);
      params.mOffImgParams.ohy = pars->get_int(csz_img_ohy);
      params.mOffImgParams.AxesUsage |= 0x2; //bits signal. 1/2/4/8/16
    }
  }
  
  //Geometry
  params.m_nShots   = pars->get_int(csz_nShots);
  params.mShot1_ix  = pars->get_int(csz_shot1_ix);
  params.mShot1_iz  = pars->get_int(csz_shot1_iz);
  params.mShotIncX  = pars->get_int(csz_shotIncX);
  
  params.mTrace1ix  = pars->get_int(csz_trace1_ix);
  params.mTrace1iz  = pars->get_int(csz_trace1_iz);
  
    
  params.mTraceIncX = pars->get_int(csz_traceIncX);
  params.m_nTraces  = pars->get_int(csz_nTraces);

  if (b3D) {
    params.mTrace1iy  = pars->get_int(csz_trace1_iy);
    params.mShot1_iy  = pars->get_int(csz_shot1_iy);
    params.m_nShotsY  = pars->get_int(csz_nShotsY);
    params.mShotIncY  = pars->get_int(csz_shotIncY);
    params.mTraceIncY = pars->get_int(csz_traceIncY);
    params.m_nTracesY = pars->get_int(csz_nTracesY);
  }

  params.bShot_offset  = pars->get_bool(csz_bShot_offset,false); //If input CMP gather
    
  //Do src Intensity balancing in RTM imaging, not used so far
  params.bDoSrcInt = pars->get_bool("bDoSrcInt",false);
  params.bDoRecvInt = pars->get_bool("bDoRecvInt",false);
  
  //Recording Snapshot params, only for 2D RTM
  params.f_shot_snap = pars->get_int(csz_snap_f,0);
  params.j_shot_snap = pars->get_int(csz_snap_j,1);
  params.n_shot_snap = pars->get_int(csz_snap_n,0);
  checkSnapRecordParams(&params);
}

} //end namespace

#endif // __WAVEPROPAGATE_PARAM_H_INCL__

