#include "stdafx.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "WEMVA.h"
#include "modelingOp.h"
#include "paddingFD.h"
#include "SR_manager.h"
#include "boundaryFD.h"
#include "dampingFunc.h"
#include "misc.h"
#include "off2ang3d.h"

using namespace Yang;
using namespace std;

//Forward modeling
void WEMVA_FD::calcData() {
  FUNC_TRACE_BEG(WEMVA_FD);
  //Here the Data is actuall the image from the velocity.
  //Init the operator, do RTM to create the prestack image
  
  //Cannot use dynamic_cast
  modelingOp* pModeling = this->m_pModeling;
  
  FDModelingParams* pFDParam = this->m_FDModelingParam;

  pModeling->initOperatorEx(pFDParam, Yang::acMigration);   
  
  //verbPrt("\r RTM Imaging from the current velocity in WEMVA module s\n");
  //Need to allocate Img0_hx here
   
  hypercube shapeOffImg;
  
  bool bDoOffset = pFDParam->mOffImgParams.bDoOffsetGather;
  VERIFY(pFDParam->b3D == m_pWEMVAparams->b3D);
  bool b3D = m_pWEMVAparams->b3D;
  if (bDoOffset) { //figure out the size of the subsurface offset image.
    get_prestack_img_cube_shape(shapeOffImg,pFDParam->mOffImgParams,
                      pFDParam->finalImageShape, pFDParam->timeParam.dtComp);
  }
  if (!b3D) { //if not doing 3D, then everything is in core.
    if (bDoOffset) {  //Do non-zero offset
      m_pWEMVAparams->Img0_hx.alloc(shapeOffImg);
    } else { //Only do zero offset
      m_pWEMVAparams->Img0_hx.alloc(pFDParam->finalImageShape);
    }
  }
  //pModeling->migration_op(&(m_pVelAnaParam->dObsCube),&(this->m_pWEMVAparams->Img0_hx));
  //To accomodate for Oneway, here change cubeData -> dmRsf_oc, because dObserved never changes.
  if (!pFDParam->bUseOnewayPropagator) {
    assert(m_pVelAnaParam->pDataObs);
    pModeling->migration_op(m_pVelAnaParam->pDataObs,&(this->m_pWEMVAparams->Img0_hx));
    VERIFY(!b3D); //Only implement 2-D case in FDTD, can be extended later.
  } else { //Use Yaxun's oneway propagator, Use the file tag as input
    VERIFY(!m_pVelAnaParam->dObsFilename.empty());
    if (!b3D) { //2D
      pModeling->migration_op(m_pVelAnaParam->dObsFilename,&(this->m_pWEMVAparams->Img0_hx));
    } else { //3D, I/O are filenames/tags both.
      VERIFY(!m_pWEMVAparams->fn_img_off.empty()); //the file name should be assigned by now.
      PRV2(m_pVelAnaParam->dObsFilename,m_pWEMVAparams->fn_img_off);
      pModeling->migration3d_op(m_pVelAnaParam->dObsFilename,m_pWEMVAparams->fn_img_off);
    }
  }
  this->m_bAngleGatherCalculated = false; //here the AngleGather has not been calculated yet.
  LOGC("Finish calculating image gather in WEMVA_FD::calcData");
  //this->m_pWEMVAparams->Img0_hx.dump("dump2");
  //LOG(finish RTM);
  FUNC_TRACE_END();
}

const cubeDataF& WEMVA_FD::getCalcData() {
  if (!m_FDModelingParam->b3D) {
//    if (m_bCalcAngleGather) { //output angle gather
//      return m_pWEMVAparams->Img0_ax_t;
//    } else {
      return m_pWEMVAparams->Img0_hx;
//    }
  } else {
    VERIFY(0);
  }
}

int WEMVA_FD::setCalcData(const cubeDataF& data) {
//  if (m_bCalcAngleGather) { //output angle gather
//    m_pWEMVAparams->Img0_ax_t = data;
//  } else {
    m_pWEMVAparams->Img0_hx = data;
//  }
  VERIFY(!m_FDModelingParam->b3D);
  return 0;
}

string WEMVA_FD::getCalcData_fn() {
  if (m_FDModelingParam->b3D) {
    //if (m_bCalcAngleGather) { //output angle gather
    //  return m_pWEMVAparams->fn_img_ang;
    //} else {
      return m_pWEMVAparams->fn_img_off;
    //}
  } else { //not for 2D currently.
    VERIFY(0);
  }
}

int WEMVA_FD::setCalcData_fn(const string& img_fn) {
  LOGC("Enter WEMVA_FD::setCalcData_fn");
  //Mv the file to my current file name
  char szcmd[512];
  //if (m_bCalcAngleGather) { //output angle gather
  //  sprintf(szcmd,"Mv %s %s", img_fn.c_str(), m_pWEMVAparams->fn_img_ang.c_str());
  //} else {
    sprintf(szcmd,"Mv %s %s -f", img_fn.c_str(), m_pWEMVAparams->fn_img_off.c_str()); //w/out -f are both fine
  //}
  //Execute command.
  prt("Exec cmd in WEMVA_FD::setCalcData_fn:%s\n", szcmd);
  VERIFY(system(szcmd)==0);
  this->m_bAngleGatherCalculated = false;
  return 0;
}

//Generate image gather.
void WEMVA_FD::postCalcData() {
  //After then calculate the angle gathers if needed
  if (m_bCalcAngleGather) {
    if (!m_bAngleGatherCalculated) {
      if(!m_pWEMVAparams->b3D) { //if do 2D only.
        this->convertAngleOffImage(m_pWEMVAparams->Img0_ax_t, m_pWEMVAparams->Img0_hx,true);
      } else { //Use biondo's program to generate 3D angle gathers, then we need to calculate from off2ang3d program.
        off2ang3d_file(*(m_pWEMVAparams->par_o2a3d),m_pWEMVAparams->fn_img_off,m_pWEMVAparams->fn_img_ang);
      }
      m_bAngleGatherCalculated = true; //already calc angle gathers.
    } else {
      prt("Warning:enter postCalcData with bAngleGatherCalculated is true\n");
    }
  }
  return;
}

//Maximize the stack power of objective function
//Lets assume the objective function is the angle stack power,
//ie the energy of the zero offset image.
float WEMVA_FD::stackPowerObjFunc() {
  FUNC_TRACE_BEG(WEMVA_FD:stackPowerObjFunc);
  //Since we have the image in Img0_hx
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  float costFunc = 0.0f;

  cubeData<float> img_hx0;
  if (!pWEMVAparam->b3D) { //if it is 2D, then everything is incore
    //Assume the initial images are ready
    VERIFY(pWEMVAparam->Img0_hx.domain().get_n123() > 0);
    
    //Get the zero offset image.
    hypercube offsetImgShape = pWEMVAparam->Img0_hx.domain();
    hypercube zeroOffImgShape = offsetImgShape.return_compatible(1,2);
    
    //Should not try to take the short cut to calculate the zero offset image. Rather I should stack over all angles. !!!!
    img_hx0.alloc(zeroOffImgShape);
    windParam wind(3,windowAxis());
    axis ax_h = offsetImgShape.get_axis(3);
    wind[2].f = ax_h.get_offset(0.0f); wind[2].n = 1;
    pWEMVAparam->Img0_hx.grab2d(&img_hx0,wind);
    
  } else { //the image is in the file name below,
    //the axis (z,x,y,hx,hy), thus take the hx=hy=0 slice.
    string fn = pWEMVAparam->fn_img_off;
    PRV(pWEMVAparam->fn_img_off);
    dmRsf_oc<float> dm_img(fn,dm::IN,"input off img",true);
    dm_img.init(NULL,5);
    windParam wd(5,windowAxis());
    axis ax_hx,ax_hy;
    ax_hx = dm_img.domain().get_axis(4); ax_hy = dm_img.domain().get_axis(5);
    wd[3].f = ax_hx.get_offset(0.0f);wd[3].n=1;
    wd[4].f = ax_hy.get_offset(0.0f);wd[4].n=1;
    img_hx0.alloc(dm_img.domain().return_compatible(1,3));
    dm_img.grabWindow(&img_hx0,wd);
    //dm_img.unInit(); //no need to call, the destructor will call it automatically.
  }
  img_hx0.dump("dump1");
  if (pWEMVAparam->bCompensateImageZ) {
    //scale the image by depth, to compensate the amplitude decay with depth.
    axis ax_z = img_hx0.domain().get_axis(1);
    float zrange = ax_z.n * ax_z.d;
    cubeData<float> cubz(hypercube(ax_z.n));
    for (int i=0; i<ax_z.n; i++) {
      cubz(i) = ax_z.get_coord(i)-ax_z.o + 0.2*zrange;  //The shallowest depth will have a scaling coef of 0.2*zrange
    }
    cubz.dump("dump4");
    multiplySlice(img_hx0,img_hx0,cubz);
  }
  
  //since this is maximization, therefore scale with -1.
  costFunc = -1.0 * dotProd(img_hx0,img_hx0);

  FUNC_TRACE_END(costFunc);
}

//FWI like WEMVA backprojection method, 3D case not finished yet!~~
void  WEMVA_maxAngleStack::calcBackProjData() {
  FUNC_TRACE_BEG(WEMVA_maxAngleStack::calcBackProjData);
  
  this->postCalcData(); //calc the angle gather at this step
  assert(m_bCalcAngleGather);
  VERIFY(m_bAngleGatherCalculated);
  
  cubeDataF& imgaxt = m_pWEMVAparams->Img0_ax_t;
  assert(!imgaxt.domain().isEmpty());
  //Stack the (z,ax,x) along the angle(ax) axis, the backProj is the spreading of the angle averaged trace.
  axis ax_x      = imgaxt.domain().get_axis(3);
  axis ax_angle  = imgaxt.domain().get_axis(2);
  axis ax_z      = imgaxt.domain().get_axis(1);
  
  cubeDataF&  bpjImgaxt = m_pWEMVAparams->backProjImg_ax;
  imgaxt.allocMemoryTo(bpjImgaxt);
  vecAxis axes; axes.push_back(ax_z); axes.push_back(ax_angle);
  hypercube sliceShape(axes);
  cubeData<float> imgSlice,imgSliceOut;
  imgSlice.becomeProxyCube(sliceShape);imgSliceOut.becomeProxyCube(sliceShape);
  windParam wind(3); wind[2].n=1;
  
  cubeDataF trcStacked;
  trcStacked.alloc(hypercube(ax_z));
  
  int ibeg=0, iend=ax_x.n;
  //If only calc sensitivity kernel, then only back project on image point of the angle gather.
  if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
    ibeg = ax_x.n/2; iend = ibeg+1;
  }
  for (int ix=ibeg; ix<iend; ix++) {
    wind[2].f = ix;
    imgaxt.setProxyXd(&imgSlice,wind,2);
    bpjImgaxt.setProxyXd(&imgSliceOut,wind,2);
    //Stack the angle
    trcStacked.zero(); stackTrc(trcStacked,imgSlice);
    spreadTrc(imgSliceOut,trcStacked);
  }
  
  if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
    cubeData<float> mask_x(ax_x);
    mask_x = 0.0f;
    //Put six spikes into the domain
    //int dnw = ax_x.n/8;
    //for (int i=0; i<7; i++) {
    //  mask_x(ax_x.n/2+(i-3)*dnw) = 1.0f;
    //}
    mask_x(ax_x.n/2) = 1.0f;
    scaleSlice(bpjImgaxt,bpjImgaxt,mask_x);
    LOG(calc only sens kernel in maxStackPowerBackProjImg);
  }
  
  imgaxt.dump("angCube"); bpjImgaxt.dump("angCubeBp");
  m_pWEMVAparams->Img0_hx.allocMemoryTo(m_pWEMVAparams->backProjImg_hx);
  this->convertAngleOffImage(bpjImgaxt,m_pWEMVAparams->backProjImg_hx,false);
  //Need to convert back to the offset image
  FUNC_TRACE_END();
}

float WEMVA_maxAngleStack::calcObjFunc() {
  return (this->stackPowerObjFunc());
}


//velGradCube is been allocated outside.
void WEMVA_FD::calcSlowGradient( cubeData<float> & slowGradCube) {
  FUNC_TRACE_BEG(WEMVA_FD::calcSlowGrad);
  WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
  FDModelingParams* TomoParams = m_FDModelingParam;
  modelingOp* pModeling = m_pModeling;
  
  
  //Init the operator, doing TomAdj this time
  pModeling->initOperatorEx(TomoParams, Yang::acTomAdj);
  
  //Verify the size of velGradCube
  slowGradCube.zero();
  //doing TomAdj this time to calculate gradient
  //The input is in &(pWEMVAparam->backProjImg_hx
  //pWEMVAparam->backProjImg_hx.dump("dump5");
  bool b3D = pWEMVAparam->b3D;
  
  
  
  if (!b3D) { //doing 2D case.
    pModeling->TomAdj_op(&(pWEMVAparam->backProjImg_hx) ,&slowGradCube);
  } else { //doing 3D case
    pModeling->TomAdj3d_op(pWEMVAparam->fn_bpjImg_off   ,slowGradCube);
    //string fn_slowGrad;
    //ostringstream oss;
    //oss << "slgd" << rand()%65536 << ".H";
    //fn_slowGrad = oss.str();
    //  //Then we read the file back into memory
    //  dmRsf_oc<float> dm_slow(fn_slowGrad,dm::IN,"",true);
    //  dm_slow.init(NULL);
    //  assert(slowGradCube.domain().sameShape(dm_slow.domain()));
    //  dm_slow.seq_grabAll(slowGradCube);
  }
  
  //Don't need to scale the slowness grad by s0, it is included in the TomAdj Operator
  //slowGradCube.dump("dump3");
  //reason of wrong sign,could not figure out so far   
  this->flipGradient(slowGradCube);
  FUNC_TRACE_END();
}

void WEMVA_FD::flipGradient(cubeData<float>& gradCube) {
   //Because the magic -1 phenomenon, flip the sign
   //gradCube.scale(-1.0f);
}

void WEMVA_FD_DSO_off::flipGradient(cubeData<float>& gradCube) {

   //gradCube.scale(-1.0f);
}

void WEMVA_FD_DSO_off::gen_h_term_cube(cubeData<float>& hTermCube, const hypercube& shapeOffImage, int h_order) {
  FUNC_TRACE_BEG(WEMVA_FD_DSO_off::gen_h_term_cube);
   int ndim = shapeOffImage.get_ndim();
   hypercube shape_h_term = shapeOffImage.return_axes(3, ndim);
   hTermCube.alloc(shape_h_term);
   //Just do 2D in this case
   assert(ndim==3);
   axis ax_hx = shape_h_term.get_axis(1);
   //Set the h cube
   float fh,h;
   for (int i=0; i<ax_hx.n; i++) {
	  h = i*ax_hx.d + ax_hx.o; fh=1.0f;
	  for (int j=0; j<h_order; j++) {
		 fh = fh*h;
	  }
	  hTermCube[i] = fh;
   }
   FUNC_TRACE_END();
}

//virtual function
float WEMVA_FD_DSO_off::calcObjFunc() {
  FUNC_TRACE_BEG(WEMVA_FD_DSO_off::calcObjFunc);
  
   //need to calcualte the initial images first somewhere before
   //Check the initial Image is not NULL
   
   WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
   VERIFY(pWEMVAparam->Img0_hx.domain().get_ndim() >= 2);

   //then calc the cost function, which is the || h*I(x,h) ||^2
   //Since we have the image in Img0_hx
   
   const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
   cubeData<float> hTermCube;
   this->gen_h_term_cube(hTermCube, shapeOffImage,1);
   pWEMVAparam->Img_hx_res.alloc(shapeOffImage);
   scaleSlice(pWEMVAparam->Img_hx_res, pWEMVAparam->Img0_hx,hTermCube);

   float costFunc = dotProd(pWEMVAparam->Img_hx_res,pWEMVAparam->Img_hx_res);
   prt("finish calc Obj Func in WEMVA_FD_DSO, obj = %f\n", costFunc);
   FUNC_TRACE_END(costFunc);
}

//Calculate Gradient, this is a virtual function
void WEMVA_FD_DSO_off::calcBackProjData() {
   //The Back Proj data is the Image -(I(x,h)h^2)
   //The Image0 is been calculated already.
   WEMVA_Param* pWEMVAparam = m_pWEMVAparams;

   const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
   cubeData<float> hTermCube;
   this->gen_h_term_cube(hTermCube, shapeOffImage,2);	//Need h^2
   pWEMVAparam->backProjImg_hx.alloc(shapeOffImage);
   scaleSlice(pWEMVAparam->backProjImg_hx, pWEMVAparam->Img0_hx,hTermCube);

   //If only calc sensitivity kernel, then only back project on image point of the angle gather.
   //i.e Mask out all the values in the image gathers but the center one.
   if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
     axis ax_x = pWEMVAparam->backProjImg_hx.domain().get_axis(2);
     cubeData<float> mask_x(ax_x);
     mask_x = 0.0f;
     
     //Put six spikes into the domain
     int dnw = ax_x.n/8;
     for (int i=0; i<7; i++) {
       mask_x(ax_x.n/2+(i-3)*dnw) = 1.0f;
     }
     
     multiplySubCube(pWEMVAparam->backProjImg_hx,pWEMVAparam->backProjImg_hx,mask_x,2,2);
   }
   
   //There is a negative sign 
   pWEMVAparam->backProjImg_hx.scale(-1.0f);
}


//angle domain DSO implementation. Taper on the edge

void WEMVA_FD_DSO_ang::flipGradient(cubeData<float>& gradCube) {
   //gradCube.scale(-1.0f);
}

float WEMVA_FD_DSO_ang::calcObjFunc() {

   //need to calcualte the initial images first somewhere before
   //Check the initial Image is not NULL
   
   WEMVA_Param* pWEMVAparam = m_pWEMVAparams;
   VERIFY(pWEMVAparam->Img0_hx.domain().get_ndim() >= 2);

   //then calc the cost function, First tranform it to angle domain, and then take derivative
   //along the angle axis, and then taper.
   //const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
   //!!! Wait for implement
   float costFunc = 0.0f;
   //prt("finish calc Obj Func in WEMVA_FD_DSO_ang, obj = %f\n", costFunc);
   return costFunc;
}


void WEMVA_FD_DSO_ang::calcBackProjData() {
  FUNC_TRACE_BEG(WEMVA_FD_DSO_off::calcBackProjData);
   //The Back Proj data is the Image -(dI^2(x,a)/da^2). 
   //The Image0 is been calculated already, Second derivative along the angle axis.
   VERIFY(m_bAngleGatherCalculated);
   
   const hypercube& shapeOffImage = m_pWEMVAparams->Img0_hx.domain();
   axis ax_off = shapeOffImage.get_axis(3);
   float hmax = -1.0f*ax_off.o; //assume the offset axis is from -hmax -> +hmax
   
   //transpose from z,x,hx --> z,hx,x and then perform off2ang tranform.
   cubeData<float>& Img0_hx_t = mImg0_hx_t;
   seplib::callSepTool(Img0_hx_t,m_pWEMVAparams->Img0_hx,"Transp","plane=23");
   
   //Transform to angle domain.
   Yang::off2Ang2d(m_pWEMVAparams->Img0_ax_t,Img0_hx_t,false,m_pWEMVAparams->off2angType,m_pWEMVAparams->amax);
   
   //Then take the 2nd derivative on angle axis.
   m_bpjImg_ax_t.alloc(m_pWEMVAparams->Img0_ax_t.domain());
   deriv2nd(m_bpjImg_ax_t,m_pWEMVAparams->Img0_ax_t,2,true);
   
   //m_pWEMVAparams->Img0_ax_t.dump("dump1");
   //m_pWEMVAparams->backProjImg_ax_t.dump("dump2");
   //Might need tapering here.
   cosDampingFunc dampFunc;
   stdVecFloat coefVec; int length = int(ax_off.n/2);
   dampFunc.getCoefEx(coefVec,length,1.0f);
   cubeData<float> TaperCube(ax_off.n);

   assert(ax_off.n%2 == 1);
   TaperCube[length] = 1.0f;
   for (int i=0; i<length; i++) {
     TaperCube[length+1+i] = coefVec[i];
     TaperCube[length-1-i] = coefVec[i];
   }
   
   //scale along the 2nd(ax) axis.
   if (m_pWEMVAparams->bTaperAngleGather) {
     multiplySubCube(m_bpjImg_ax_t,m_bpjImg_ax_t,TaperCube,2,2);
   }
   
   cubeData<float> backProjImg_hx_t;
   //Then tranform back to offset domain
   Yang::off2Ang2d(backProjImg_hx_t,m_bpjImg_ax_t,true,
                   m_pWEMVAparams->off2angType,hmax);
                                                                                         
   //Transpose back
   seplib::callSepTool(m_pWEMVAparams->backProjImg_hx, backProjImg_hx_t,"Transp","plane=23");
   
   //There is a negative sign 
   m_pWEMVAparams->backProjImg_hx.scale(-1.0f);
  FUNC_TRACE_END();
}

//For WEMVA objFunc to update status in a iterated solver
void WEMVA_FD::callBackAtIter(int iIter) {
  //update my status, currently just change the dJdu param
  bool bUseSembForDJDU = m_pVelAnaParam->bUseSemblanceToCalcDJdb;
  if (bUseSembForDJDU) {
    const axis &ax_semb = m_pVelAnaParam->ax_semb;
    float fMinSembWnd = ax_semb.d * 6.0f; 
    float fSembWindow = m_pVelAnaParam->fSembSmoothWindow;
    m_pVelAnaParam->fSembSmoothWindow = MAX(fSembWindow/1.3f, fMinSembWnd);
  }
  return;
}

/////////////////// WEMVA objective function Interfaces///////////////////////////////////////////
//model is velocity, simply wrap around the function interface
float WEMVA_FD::evaluate(const gVecComponent& model) {
  FUNC_TRACE_BEG(WEMVA_FD::evaluate);
  //set the model to the FWI object
  //convert slowness to velocity
  cubeData<float> vel;
  //LOG(mm1);
  model.allocMemoryTo(vel);
  Yang::reciprocal(vel,model);
  //LOG(mm2);
  this->setVelocity(&vel);
  
  //LOG(calcualte initial image);
  //First calculate the data
  this->calcData();
  
  //LOG(mm4);
  //LOG(calc obj func in wemva_evaluate);
  float ret = this->calcObjFunc();
  if (std::isnan(ret)) {
    prt("\n the cost func is nan ! ");
  }
  FUNC_TRACE_END(ret);
}

void  WEMVA_FD::calcSearchDir( gVecComponent &searchDir) {
  FUNC_TRACE_BEG(WEMVA_FD::calcSearchDir);
  //Here we have alrceady calculated the data, since we evaluated the obj function already.
  this->calcBackProjData();
  
  //calculate gradient this time
  //It is actually SLOWNESS gradient
  this->calcSlowGradient(searchDir);
  
  //m_pWEMVA->m_pWEMVAparams->Img0_hx.dump("dump6");
  //searchDir.dump("dump7");
  FUNC_TRACE_END();
}

