#ifndef __WEMVA_H_INCL__
#define __WEMVA_H_INCL__

// TODO: your code here
#include "velAna.h"
#include "objFunc.h"

namespace Yang {

  //Inherit the objFunc interface as well 
class WEMVA_FD:public velAna_FD, public objFunc {
public:
	// Constructor
	WEMVA_FD(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar):
			 velAna_FD(pVelAnaParam, pFDModelingParam, pModel),m_pWEMVAparams(pPar), m_bCalcAngleGather(true), m_bAngleGatherCalculated(false){}
	// Destructor
	virtual ~WEMVA_FD();

	virtual void calcData();
  void postCalcData();
	virtual void calcBackProjData() = 0;
	void calcSlowGradient(cubeData<float>& velGradCube);
	
	//calculate objective function, used in solver
	virtual float calcObjFunc() = 0;
  
  void convertAngleOffImage(cubeData<float>& img_ax_t, cubeData<float>& img_hx, bool bOff2Ang);

  virtual const cubeDataF& getCalcData();
  virtual string getCalcData_fn();
  virtual int   setCalcData(const cubeData<float>& data);
  virtual int   setCalcData_fn(const string& fn_data);
  
  //The interfaces from the objFunc class
  virtual float evaluate(const gVecComponent& model);
  virtual void  calcSearchDir( gVecComponent& searchDir);
  virtual void  callBackAtIter(int iIter); //call back for iterative solver
  
protected:
	virtual void flipGradient(cubeData<float>& gradCube);
  float stackPowerObjFunc();
  void spreadWtFromCoarseToFineGrid(cubeDataF& wtCube,const cubeDataF& coarseCube);
  void findSmoothDJDUfromSemb(cubeDataF& dJdu, cubeDataF& sembStack, int iSembAxis);
public:
	WEMVA_Param *m_pWEMVAparams;

  //intermediate memory objects;
  cubeData<float> mImg0_hx_t;
  cubeData<float> m_bpjImg_ax_t;

public: //protected:
  bool m_bCalcAngleGather;
  bool m_bAngleGatherCalculated;
};

// Destructor implementation
inline WEMVA_FD::~WEMVA_FD()
{
}

//offset domain implementation
class WEMVA_FD_DSO_off : public WEMVA_FD {
public:
  WEMVA_FD_DSO_off(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar): 
	WEMVA_FD(pVelAnaParam, pFDModelingParam,pModel,pPar) {m_bCalcAngleGather=false;}
   
   virtual void calcBackProjData();
   virtual float calcObjFunc();
protected:
	virtual void flipGradient(cubeData<float>& gradCube);
	void gen_h_term_cube(cubeData<float>& hTermCube, const hypercube& shapeOffImage, int h_order);
};

//angle domain implementation
class WEMVA_FD_DSO_ang : public WEMVA_FD {
public:
   WEMVA_FD_DSO_ang(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar): 
	 WEMVA_FD(pVelAnaParam, pFDModelingParam,pModel,pPar) {}
   
   virtual void calcBackProjData();
   virtual float calcObjFunc();
protected:
	virtual void flipGradient(cubeData<float>& gradCube);
};
  
//Maximize stack power
class WEMVA_maxAngleStack: public WEMVA_FD{
public:
  WEMVA_maxAngleStack(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
       modelingOp* pModel, WEMVA_Param * pPar): 
  WEMVA_FD(pVelAnaParam, pFDModelingParam,pModel,pPar) {}
  
  virtual void calcBackProjData();
  //calculate objective function, used in solver
  virtual float calcObjFunc();
};

class WEMVA_FD_eventBased : public WEMVA_FD {
public:
  WEMVA_FD_eventBased(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar):
	WEMVA_FD(pVelAnaParam, pFDModelingParam,pModel,pPar) {}
   
  virtual void calcBackProjData();
  virtual void calcBackProjData2d() {}
  virtual void calcBackProjData3d() {}
  virtual float calcObjFunc() = 0;
protected:
};

//RMO style, tan(\gamma)^2 moveout function or (two params (plus a shift) for each image pt.
class WEMVA_FD_beamFocus : public WEMVA_FD_eventBased {
public:
   WEMVA_FD_beamFocus(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar): 
	 WEMVA_FD_eventBased(pVelAnaParam, pFDModelingParam,pModel,pPar) {}
   
   virtual float calcObjFunc();
   //void calcBackProjData();

   virtual void getfrCube(cubeData<float>& frCube);
   void calcDJDU();
   virtual void calcDJDU2d();
   virtual void calcDJDU3d();
   
   virtual void calcBackProjData2d();
   virtual void calcBackProjData3d();
protected:
	virtual void flipGradient(cubeData<float>& gradCube);
  void calcBackProjData3d_mJacob();
  void calcBackProjData3d_2Jacob();
};

//Use curvature for RMO, but the deltaS is linked to the the shifts of individual traces
class WEMVA_FD_beamFocus_m: public WEMVA_FD_beamFocus {
public:
   WEMVA_FD_beamFocus_m(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar): 
	 WEMVA_FD_beamFocus(pVelAnaParam, pFDModelingParam,pModel,pPar) {}
   
   //The only difference to beamFocus is how to calculate du/ds
   virtual void calcBackProjData2d();
protected:
	 virtual void flipGradient(cubeData<float>& gradCube) {}
};

//Use non-picking, correlation based objective function, use zero angle as reference.Based on angle moveout-parameters
class WEMVA_FD_angleMoveout : public WEMVA_FD_eventBased {
public:
   WEMVA_FD_angleMoveout(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar): 
	 WEMVA_FD_eventBased(pVelAnaParam, pFDModelingParam,pModel,pPar) {}
   
   virtual float  calcObjFunc();
   virtual void   calcBackProjData2d();
protected:
  void calcBackProjDataMultipleEvents(cubeDataF& imgAngleFlat);
};


//Use non-picking, correlation based objective function, use time shift images.
class WEMVA_FD_beamFocusTime : public WEMVA_FD_eventBased {
public:
   WEMVA_FD_beamFocusTime(velAnaParam* pVelAnaParam, FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel, WEMVA_Param * pPar): 
	 WEMVA_FD_eventBased(pVelAnaParam, pFDModelingParam,pModel,pPar) {m_bCalcAngleGather=false;}
   
   virtual float calcObjFunc();
   virtual void calcBackProjData2d();
protected:
};

}

#endif // __WEMVA_H_INCL__

