#ifndef __VELANAPARAMS_H_INCL__
#define __VELANAPARAMS_H_INCL__

#include "cubeData.hpp"
#include "cubeAlgo.h"
#include "modelingFD_aux.h"
#include "modelingOp.h"
#include "imaging.h"
#include "cubeAlgo2.h"

// TODO: your code here
namespace Yang {
struct velAnaParam {
   cubeData<float> velGradCube;
   cubeData<float> velModelMask; //Mask the part of vel model that can be resolved by survey geometry

   
   string dObsFilename;
   dmRsf<float>* 	  pDataObs;
   cubeData<float>  dObsCube; //in core observed data, not always in place
   
   //For Beamfocusing
   int beamSize;
   //Currently quadratic form, theta= u*(offset)^2
   //assume input is 1d, axis is along offset
   tagMoveOutType moveOutType;   
      
	//For Yang's Beam focusing
  bool bIgnoreBfY2ndBpjTerm;
	cubeData<float> pickedCurvatureCube; //a param
	cubeData<float> pickedSlopeCube; // b parameter 
  //Subsampled version, because picking is kind of expensive.
  cubeData<float> pickedCurvatureCubeSubsamp; 
	cubeData<float> pickedSlopeCubeSubsamp; 
	axis ax_a, ax_b; //axes for a,b parameter scaning.

  cubeData<float> dJdbCube; //The derivative of obj function over moveout params
  bool  bUseSemblanceToCalcDJdb;
  cubeData<float> semblanceCube;
  axis ax_semb;
  bool  bNormalizeSemb;
  bool bUseGuassianDerivForDJDU; float fSembSmoothWindow;
  
  typedef enum tagMaxCorrUsageType {
    pickMaxCorr,
    smoothCorrDeriv0,
    GaussianWeightCorr
  } maxCorrUsageType;
  maxCorrUsageType maxCorrUsage;
  
  bool bPickMoveoutPar; //if false, use derivative at zero
  
  bool bDivideEnergyTermUseMask;
  //default param values
	velAnaParam(): beamSize(0), bIgnoreBfY2ndBpjTerm(false), bUseSemblanceToCalcDJdb(false),bDivideEnergyTermUseMask(false) {}
};

//Memory objects and params used in FWI or Transmission (Data space inversion)
//public velAna_FD_Param 
struct FWI_Param {
  typedef enum tagObjFuncType {
    classic,
    YiLuo,
    YiLuo2,
    beamFocusing1,  //single param fitting
    beamFocusing2,  //double param fitting
    beamFocusing21, //double param, but with shift and slope
    beamFocusingY1,	//require picking, Just measuring the slope
    beamFocusingY2,	//require picking, 2 parameter picking, but we can also use the slope only, 
                    //by setting the a cube equal to zero, as the one above
    DSO //analogy to DSO
  } FWIobjFuncType;

  FWIobjFuncType mFWItype;
  
   cubeData<float> dCalCube;
   cubeData<float> backProjCube;   
   //For correlation based method
   cubeData<float> C_cube;
   cubeData<float> deltaTauCube; //shift for Maximum correlations
   int iStartLag;
   int iEndLag;
   
   //For YiLuo 1991 method
   cubeData<float> dObs2_shiftedCube;
   cubeData<float> dObs_shiftedCube;
   cubeData<float> E_cube; //scalar for each xs,xg;
   
   float widthGaussDeriv;
};

class param_off2ang3d;
//Memory objects and params used in WEMVA Tomograpy case (reflection case)
struct WEMVA_Param {
  typedef enum tagObjFuncType {
    classic,
    YiLuo,
    beamFocusing1,  //single param fitting
    beamFocusing2,  //double param fitting
    DSO_off, //offset domain DSO
    DSO_ang  //angle domain DSO
  } WEMVAobjFuncType;
   
  WEMVAobjFuncType mWEMVAtype; 
  
  enum dRho_dsOpType {drds_Jacob_2, drds_Jacob_m}; //drho/ds operator type, 1) invert 2X2 Jacobian 2) invert m+1 by m+1 Jacobian
  dRho_dsOpType m_dRho_dsOpType;
  
  //if for 3D, we need out-of-core access.
  bool b3D;
  param_off2ang3d *par_o2a3d,*par_a2o3d;
  string fn_img_off,fn_bpjImg_off;  //(z,x,y,hx,hy)
  string fn_img_ang,fn_bpjImg_ang;  //(z,gamma,azim,x,y)
  
  string fn_dJdu,fn_semb,fn_envSemb;
  //dmRsf<float>  *initImg_off_oc, *initImg_ang_oc, *backProjImg_off_oc, *backProjImg_ang_oc;
  
  
   //Initial images.
   cubeData<float> Img0;
   cubeData<float> Img0_hx; //offset image
  
   cubeData<float> backProjImg; //zero-subsurf-offset Img;
   cubeData<float> backProjImg_hx; //offset Image for back propagation.
   cubeData<float> backProjImg_ax; //angle domain bpj image
   //For angle gather offset transform
   enumOff2angType off2angType;
   int ax_ang_n; //The number of points on angle axis
     
   //Parameters defining the range of h when imaging I(x,h)
   offsetImagingParams off_img_param;
   //For DSO    
   cubeData<float> Img_hx_res; //Img residue. 
   //For YiLuo 1991 method
   
   //For beam Focus Biondi,sep142
   //These images axises are transposed. e.g (z,ax,x) Not (z,ax,x)
   cubeData<float> Img0_ax_t; //The offset/angle domain Image, transposed
   // cubeData<float> dJduCube; //calc the derivative of J over the moveout params. (z,x), already defined in its father class
   cubeData<float> Ecube,finalWtCube; //(z,x), the denumerator

   float thresMaxValPercent; //from 0~1
   float fEmaskValThres;  //the alternative, an arbitrary value.
   cubeData<float> thresMaskCube;

   bool bSingleEvent; //If there is only one event
   bool bUseSingleMoveoutParam; //If use only one moveout params.
   bool bOnlyCalcSensitivityKernel; //if then just set dJ/du = spike.
   float amax; //max angle for angle transform
   
   //Taper angle gathers for far angle, used in DSO.
   bool bTaperAngleGather;
   
   axis ax_beam_z; //For dividing windows along z directions
   float beam_z_window;
   hypercube imageShape; //zero offset image.
   bool bBalanceEnumeratorWt; //Will normalize the amplitude of the F matrix in beamfocus method.
   
   //For correlation based shift parameter method.
   bool bFindShiftUserCorr;
   int iStartLag,iEndLag; //[istart,iend]
   //For the method using LS fitting to determine curvature change to shift at each angle
   bool bUseWeightedLSfit;
   bool bNoLSfit;
   
  bool bCompensateImageZ; //compensate amplitude along depth direction of migrated image.
   WEMVA_Param(): par_o2a3d(NULL), thresMaxValPercent(0.1), bSingleEvent(true), bUseSingleMoveoutParam(true),
     bOnlyCalcSensitivityKernel(false),
     amax(75.0f), bTaperAngleGather(false),bFindShiftUserCorr(false),iStartLag(0),iEndLag(0), bCompensateImageZ(true),
     m_dRho_dsOpType(drds_Jacob_m)
     //initImg_off_oc(NULL),initImg_ang_oc(NULL),backProjImg_off_oc(NULL),backProjImg_ang_oc(NULL)
   {}
};

class velAna{
public:
   //Data objects for vel Ana
   velAnaParam* m_pVelAnaParam;
public:
   velAna(velAnaParam* pVelAnaParam): m_pVelAnaParam(pVelAnaParam) {}
   static void setMoveoutDeriv(cubeData<float>& cube, tagMoveOutType tp);
};

//This is the base class for Time Finite difference velocity analysis method.
class velAna_FD:public velAna {
public:
	  FDModelingParams* m_FDModelingParam; //this is for Tomo
	  modelingOp* m_pModeling;
public:
	// Constructor
	velAna_FD(velAnaParam* pVelAnaParam,FDModelingParams* pFDModelingParam, 
			  modelingOp* pModel): velAna(pVelAnaParam), m_FDModelingParam(pFDModelingParam), 
			  m_pModeling(pModel) {}

	// Destructor
	virtual ~velAna_FD();
	
	virtual void loadDataObs(); //should not be inherited anymore
	virtual void setVelocity(const cubeData<float> * pVelCube);
	
	virtual void calcData() = 0;
	virtual void calcBackProjData() = 0;
	virtual void calcSlowGradient(cubeData<float>& velGradCube) = 0;
	
	//calculate objective function, used in solver
	virtual float calcObjFunc() = 0;

  virtual void callBackAtIter(int iIter) {} //Call back function for a nonlin solver
protected:
	virtual void flipGradient(cubeData<float>& gradCube) = 0;
};

// Destructor implementation
inline velAna_FD::~velAna_FD()
{
}

}

#endif // __VELANAPARAMS_H_INCL__

