#include "stdafx.h"
#include "param_func.h"
#include "velAna.h"
#include "WEMVA.h"

#include "FWI_FD.h"
#include "FWI_FD_beamFocus.h"

#include "velAna.param.h"
#include "param_entry.h"

// TODO: your code here
using namespace Yang;

void Yang::read_velAna_param(velAnaParam* pParam, param_func* pars) {
    velAnaParam& velAnaParams = *pParam;
    //beam focus related
    velAnaParams.beamSize = pars->get_int(csz_subGroupSize,-1); //e.g. beam size;
  
    int moveOutFuncType = pars->get_int("moveOutType",2);
    switch(moveOutFuncType) {
      case 1: 
      velAnaParams.moveOutType = Linear;
      break;
      case 2:
      velAnaParams.moveOutType = Quadratic;
      break;
      case 3:
      velAnaParams.moveOutType = tangentSquare;
      default:
      seperr("wrong moveOutType for beam focusing");
      break;
    }

    velAnaParams.bUseSemblanceToCalcDJdb = pars->get_bool("bUseSemblance",false);
    if (velAnaParams.bUseSemblanceToCalcDJdb) { //read the axis for semblance scan
      velAnaParams.ax_semb.d = pars->get_float("ax_semb_d");
      velAnaParams.ax_semb.n = pars->get_int("ax_semb_n");
      float ax_semb_o = (velAnaParams.ax_semb.n/2)*-1.0f*velAnaParams.ax_semb.d;
      velAnaParams.ax_semb.o = pars->get_float("ax_semb_o",ax_semb_o);
      velAnaParams.ax_semb.print_info("ax_semb");
      velAnaParams.bUseGuassianDerivForDJDU   = pars->get_bool("bUseGaussianDerivForDJDU", false);
      PRV(velAnaParams.bUseGuassianDerivForDJDU);
      velAnaParams.fSembSmoothWindow          = pars->get_float("fSembSmoothWindow",  velAnaParams.ax_semb.n*velAnaParams.ax_semb.d/2.0f);
      velAnaParams.bNormalizeSemb             = pars->get_bool("bNormSemb",false); 
      PRV(velAnaParams.bNormalizeSemb);
    }
    
    int indexCorrUsage = pars->get_int("corrUsage",0);
    switch(indexCorrUsage) {
    case 0:
      velAnaParams.maxCorrUsage = velAnaParam::pickMaxCorr;       break;
    case 1:
      velAnaParams.maxCorrUsage = velAnaParam::smoothCorrDeriv0;  break;
    case 2:
      velAnaParams.maxCorrUsage = velAnaParam::GaussianWeightCorr;break;
    default:
      VERIFY(0);
    }
    velAnaParams.bPickMoveoutPar= pars->get_bool("bPickMoveoutPar",false);
    if (velAnaParams.bPickMoveoutPar) {
      PRV(velAnaParams.bPickMoveoutPar);
      VERIFY(velAnaParams.bUseSemblanceToCalcDJdb);
    }
    
    velAnaParams.bDivideEnergyTermUseMask = pars->get_bool("bDivideUseMask",true);
    return;
  }
  
  void Yang::read_WEMVA_param(WEMVA_Param* pParam, param_func* pars) {
    FUNC_TRACE_BEG(read_WEMVA_param);
    WEMVA_Param& WEMVAparam = *pParam;
    WEMVAparam.bSingleEvent = pars->get_bool(csz_bSingleEvent,true);
    WEMVAparam.bTaperAngleGather = pars->get_bool(csz_bTaperAngGather,false);

    WEMVAparam.amax = pars->get_float(csz_amax,50.0f); //if use two way prop, then the amax could be higher.
    
    WEMVAparam.bUseSingleMoveoutParam = pars->get_bool("bUseSingleMoveoutParam",false);
    WEMVAparam.bOnlyCalcSensitivityKernel = pars->get_bool("bSensKernel",false);
    int off2angTypeIndex = pars->get_int(csz_off2angType, 0);
    switch (off2angTypeIndex) {
      case 0:
        WEMVAparam.off2angType = Yang::slantstackYangAA;
        break;
      case 1:
        WEMVAparam.off2angType = Yang::slantstackYang;
        break;
      case 2:
        WEMVAparam.off2angType = Yang::seplibModified;
        break;
      case 3:
        WEMVAparam.off2angType = Yang::biondi3d_fd;
        VERIFY(pParam->b3D);
        break;
      default:
        VERIFY(0);
    }
    PRV(off2angTypeIndex);
    
    //The two below are alternative to each other.
    WEMVAparam.thresMaxValPercent = pars->get_float(csz_thresMax,0.05f);
    WEMVAparam.fEmaskValThres = pars->get_float(csz_thresMax,0.01f);

    //Param for offset->angle transform
    offsetImagingParams& offParam = WEMVAparam.off_img_param;
    VERIFY(offParam.nhx>1);
    int ax_ang_n = pars->get_int("ax_ang_n", offParam.nhx); //default the number of pts is the same as in offset axis
    //PRV(ax_ang_n);
    WEMVAparam.ax_ang_n = ax_ang_n;
    //Read params for multiple event beam focusing
    VERIFY(!WEMVAparam.imageShape.isEmpty());
    axis ax_imgz = WEMVAparam.imageShape.get_axis(1);
    float dz = ax_imgz.d;
    
    if (!WEMVAparam.bSingleEvent) {
      float fBeam_z_window      = pars->get_float("beam_z_window");
      int   iBeam_z_window      = lroundf(fBeam_z_window/dz);
      fBeam_z_window            = iBeam_z_window*dz;
      WEMVAparam.beam_z_window  = fBeam_z_window;
  
      float fBeam_z_interv  = pars->get_float("beam_z_interv",dz);
      int   iBeam_z_interv  = lroundf(fBeam_z_interv/dz);
      fBeam_z_interv        = iBeam_z_interv * dz;
      float fBeam_z_o       = pars->get_float("beam_z_o",  ax_imgz.o-0.5f*dz+fBeam_z_interv*0.5f);
      //put the origin to closest 0.5dz grid
      fBeam_z_o = ax_imgz.o + 0.5f*dz*lroundf((fBeam_z_o-ax_imgz.o)/(0.5f*dz));
      
      int   iBeam_z_n       = pars->get_int("beam_z_n",    int(ax_imgz.n/iBeam_z_interv));
      WEMVAparam.ax_beam_z.label = "beam_z";
      WEMVAparam.ax_beam_z.o= fBeam_z_o;
      WEMVAparam.ax_beam_z.d= fBeam_z_interv;
      WEMVAparam.ax_beam_z.n= iBeam_z_n;
      WEMVAparam.ax_beam_z.print_info("ax_beam_z");
      
      //VERIFY the axis info 
      //fBeam_z_o either sits on grid or sits in 50% between two grids, corresponding to fBeam_z_window being odd or even 
      float   coordBeam_z_o       = ax_imgz.get_offsetf(fBeam_z_o);
      bool bOddGrid = false;
      int     iCoordBeam_z_o      = int(coordBeam_z_o);
      if (fabs(coordBeam_z_o - iCoordBeam_z_o) < 0.1f) {
        bOddGrid = false;
        VERIFY(iBeam_z_window %2 == 1);
      } else if (fabs(coordBeam_z_o/dz - iCoordBeam_z_o - 0.5f) < epsilon) {
        bOddGrid = true;
        VERIFY(iBeam_z_window %2 == 0); //beam window should have even size if beam_o sits 50% grid
      }
      prt("\n iBeam_z_window=%d, iBeam_z_interv=%d", iBeam_z_window, iBeam_z_interv);
      if (iBeam_z_window % iBeam_z_interv != 0) {
        prt("\n WARN: iBeam_z_window % iBeam_z_interv !!== 0, the beam window does not cover the image point uniformly.");
      }
      
      //Check the beam window, should not be out of range.
      float max_z; ax_imgz.getRange(NULL,&max_z);
      VERIFY(fBeam_z_o - ax_imgz.o - fBeam_z_window*0.5f + 0.5f*dz >= 0.0f);
      VERIFY(fBeam_z_o + (iBeam_z_n-1)*fBeam_z_interv + (fBeam_z_window-dz)*0.5f <= max_z);
      
      fBeam_z_window -= dz; //subtract dz for the convenience of calculating the start/end grid point
      WEMVAparam.beam_z_window = fBeam_z_window; 
      PRV(WEMVAparam.beam_z_window);
    }
    
    WEMVAparam.bBalanceEnumeratorWt = pars->get_bool("bBalanceEnum",false);
    //For angle stack reference shift method
    WEMVAparam.bFindShiftUserCorr = pars->get_bool("bUseCorr",true);
    WEMVAparam.iStartLag          = pars->get_int("startLag",-50);
    WEMVAparam.iEndLag            = pars->get_int("endLag",50);
    
    WEMVAparam.bUseWeightedLSfit  = pars->get_bool("bUseWeightedLSfit",true);
    WEMVAparam.bNoLSfit           = pars->get_bool("bNoLSfit",false);
    
    bool b3D = WEMVAparam.b3D;
    if (b3D) { //we should have all file names for the out-of-core results.
      WEMVAparam.fn_img_off     = pars->get_string("fImgOff");
      WEMVAparam.fn_img_ang     = pars->get_string("fImgAng");
      WEMVAparam.fn_bpjImg_off  = pars->get_string("fBpjImgOff");
      WEMVAparam.fn_bpjImg_ang  = pars->get_string("fBpjImgAng");
      
      WEMVAparam.fn_dJdu        = pars->get_string("fDJDU");
      WEMVAparam.fn_semb        = pars->get_string("fSemb");
      WEMVAparam.fn_envSemb     = pars->get_string("fEnvSemb");
    }
    FUNC_TRACE_END();
  }
  
  WEMVA_FD* Yang::create_WEMVA_obj(velAnaParam* pVApar, FDModelingParams* pModelPar, modelingOp* pModelingObj, WEMVA_Param* pWEMVApar, param_func* pars) {
    WEMVA_FD* pWEMVAobj     = NULL;
    WEMVA_Param& WEMVAparam = *pWEMVApar;
    int objFuncIndex = pars->get_int("objFunc",1);
     //create WEMVA FWIObj
     switch(objFuncIndex) {
     case 1: 	  
       prt("\n using maximize angle stack obj function");
       WEMVAparam.mWEMVAtype = WEMVA_Param::classic; 
       pWEMVAobj = new WEMVA_maxAngleStack(pVApar, pModelPar, pModelingObj, &WEMVAparam);
       break;
     case 4:	  
      prt("\nusing minimization objective function with shift moveout params, with respect"
          " to the shift refered to the angle-stack trc");
      WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1; 
      pWEMVAobj = new WEMVA_FD_angleMoveout(pVApar, pModelPar, pModelingObj, &WEMVAparam);
      break;
     case 5:	  
      prt("\nusing img-space beam focusing");
      WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1; 
      pWEMVAobj = new WEMVA_FD_beamFocus(pVApar, pModelPar, pModelingObj, &WEMVAparam);
      break;
     case 55:
      prt("\nusing beamfocus_m method");
      WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1; 
      pWEMVAobj = new WEMVA_FD_beamFocus_m(pVApar, pModelPar, pModelingObj, &WEMVAparam);
      break;
     case 6: 
      prt("using DSO Refl obj Function, offset domain");
      WEMVAparam.mWEMVAtype = WEMVA_Param::DSO_off; 
      pWEMVAobj = new WEMVA_FD_DSO_off(pVApar, pModelPar, pModelingObj, &WEMVAparam);
      break;
     case 7: 
      prt("using DSO Refl obj Function, angle domain");
      WEMVAparam.mWEMVAtype = WEMVA_Param::DSO_ang; 
      pWEMVAobj = new WEMVA_FD_DSO_ang(pVApar, pModelPar, pModelingObj, &WEMVAparam);
      break;
     case 11:
      prt("\n using time-shift image bfocusing");
      WEMVAparam.bSingleEvent = false;
      WEMVAparam.mWEMVAtype = WEMVA_Param::beamFocusing1; 
      pWEMVAobj = new WEMVA_FD_beamFocusTime(pVApar, pModelPar, pModelingObj, &WEMVAparam);
      break;
     //case 9: break;
     //case 10: break;
     default:
      seperr("wrong  objFunction index");
      exit(-1);
     }
     return pWEMVAobj;
  }
  
  void Yang::read_FWI_param(FWI_Param *pFWIparam, param_func* pars) {
    FWI_Param& FWIparam = *pFWIparam;
    //Load the corrlation window
    FWIparam.iStartLag = pars->get_int(csz_corrStartLag);
    FWIparam.iEndLag = pars->get_int(csz_corrEndLag);
    return;
  }
  
  FWI_FD* Yang::create_FWI_obj(velAnaParam* pVApar, FDModelingParams* pModelPar, modelingFDgen_op* pModelingObj, FWI_Param* pFWIpar, param_func* pars) {
    bool bTaperEdge = true; //Has to be true!
    int objFuncIndex = pars->get_int("objFunc",1);
    FWI_FD* pFWIObj;
    FWI_Param& fwiPar = *pFWIpar;
    switch(objFuncIndex) {
     case 1: 
  	  prt("using classical FWI obj Function");
  	  fwiPar.mFWItype = FWI_Param::classic; 
  	  pFWIObj = new FWI_FD_classic(pVApar,pModelPar,pModelingObj,pFWIpar);
  	  break;
     case 2: 
  	  prt("using YiLuo FWI obj Function");
  	  fwiPar.mFWItype = FWI_Param::YiLuo; 
  	  pFWIObj = new FWI_FD_YL(pVApar,pModelPar,pModelingObj,pFWIpar);
  	  break;
     case 22: 
      prt("using YiLuo2 FWI obj Function");
      fwiPar.mFWItype = FWI_Param::YiLuo2; 
      pFWIObj = new FWI_FD_YL2(pVApar,pModelPar,pModelingObj,pFWIpar);
      break; 
     case 3: 
  	  prt("using beam focusing1 FWI obj Function\n");
  	  fwiPar.mFWItype = FWI_Param::beamFocusing1;
  	  pFWIObj = new FWI_FD_beamFocus1(pVApar,pModelPar,pModelingObj,pFWIpar,bTaperEdge);
      break;
     default:
  	  seperr("wrong  objFunction index\n");
  	  exit(-1);
    }
    return pFWIObj;
  }



